#include <iostream>
#include <sstream>
#include <fstream>
#include <vector>
#include <iomanip>
#include <cmath>
#include <assert.h>

using namespace std;

typedef double T43Matrix[4][3];
const char kBlankChars[] = " \t\n\r";

class prediction
{
    public:
        double T[12];
        double SCORE;

        prediction(double* f, double b)
        {
            for(int i=0;i<12;i++)
            {
                T[i] = f[i];
            }
            SCORE = b;
        }
        ~prediction(){}
};

#ifndef _ATOM_H_
#define _ATOM_H_

class atom
{
    public:
        int         anum;   // id of the atom
        string      atype;  // type of the atom
        int         rnum;   // residue id
        double      axyz[3]; // xyz coordinates
        string      chain; // chain id
        string      residue; // name of residue

        atom(int m_num, string m_type, int m_rid, double* m_xyz, string m_residue,
            string m_chain)
        {
            anum = m_num;
            rnum = m_rid;
            atype = m_type;
            for(int i=0;i<3;i++)
            {
                axyz[i] = m_xyz[i];
            }
            residue = m_residue;
            chain = m_chain;
        }

        atom(){}
        ~atom(){}
};

#endif

/// Returns a string with leading/trailing characters of a set stripped
string trimmed(string const& str, char const* sepSet=kBlankChars)
{
    string::size_type const first = str.find_first_not_of(sepSet);
    return (first==string::npos )? string():str.substr(first, str.find_last_not_of(sepSet)-first+1);
}

string rtrimmed(string const& str, char const* sepSet)
{
    string::size_type const last = str.find_last_not_of(sepSet);
    return (last==string::npos)? string():str.substr(0, last+1);
}

string ltrimmed(string const& str, char const* sepSet)
{
    string::size_type const first = str.find_first_not_of(sepSet);
    return (first==string::npos)? string():str.substr(first);
}


void read_protein(string infile, vector<atom>& pdbdata)
{
    ifstream ifs;
    ifs.open(infile.c_str());
    if(!ifs)
    {
        cerr << "Could not open file: " << infile << endl;
        exit(EXIT_FAILURE);
    }

    string line;

    int atom_id, res_id;
    string chain_id;
    string res_name, atom_type, achar;
    double d_xyz[3];

    vector<atom> vct_atom;

    while(getline(ifs, line))
    {
        if((line.substr(0, 4)).compare("ATOM") == 0)
        {
            atom_id = atoi((line.substr(6, 5)).c_str());
            atom_type = trimmed(line.substr(12, 5), kBlankChars); // hydrogen atoms have 4 letter identification
            res_name = trimmed(line.substr(17, 3), kBlankChars);
            chain_id = trimmed(line.substr(21, 1), kBlankChars);
            res_id = atoi((line.substr(22, 4)).c_str());
            //cout << res_id << endl;
            d_xyz[0] = atof((line.substr(30, 8)).c_str());
            d_xyz[1] = atof((line.substr(38, 8)).c_str());
            d_xyz[2] = atof((line.substr(46, 8)).c_str());

            // create atom object
            atom n_atom(atom_id, atom_type, res_id, d_xyz, res_name, chain_id);

            vct_atom.push_back(n_atom);
        }
    }// end while
    ifs.close();
    
    pdbdata = vct_atom;
}

void read_predictions(string infile, vector<prediction>& Z, double* transL, double* rotL)
{
    ifstream ifs;
    ifs.open(infile.c_str());
    if(!ifs)
    {
        cerr << "Could not open file: " << infile << endl;
        exit(EXIT_FAILURE);
    }

    string line;
    rotL[0]=rotL[1]=rotL[2]=transL[0]=transL[1]=transL[2]=0;

    double t[12], sc;
    while(getline(ifs, line))
    {
        if(line.find("LIG:") != string::npos)
        {
            istringstream iss(line);
            string s;
            iss >> s >> rotL[0] >> rotL[1] >> rotL[2] >> transL[0] >> transL[1] >> transL[2];
            continue;
        }

        vector<double> V;
        istringstream iss(line);
        for(int i=0;i<12;i++) {
           iss >> t[i];
	   // cout << t[i] << " ";
	}
        iss >> sc;
	assert(!iss.fail());
	// cout << sc << endl;

        prediction np(t, sc);
        Z.push_back(np);
    }
    ifs.close();

    if(Z.size() == 0)
    {
        cerr << "Error: no data to process" << endl;
        exit(EXIT_FAILURE);
    }
}

bool scoresort(const prediction& a, const prediction& b)
{
    return a.SCORE > b.SCORE;
}

string to_string(int t)
{
    stringstream ss;
    ss << t;
    return ss.str();
}

void print_complex(string fname, int complex_id, vector<atom>& L)
{
    string pdb_file_name = fname + to_string(complex_id) + ".pdb";
    FILE        *pdb_file ;
    if((pdb_file = fopen( pdb_file_name.c_str(), "w")) == NULL)
    {
        printf( "This file could not be opened.\nDying\n\n" ) ;
        exit(EXIT_FAILURE);
    }

    for(size_t i=0;i<L.size();i++)
    {
        fprintf(pdb_file, "ATOM  %5d  %-4s%3s %1s%4d    %8.3f%8.3f%8.3f\n",
            L[i].anum, L[i].atype.c_str(), L[i].residue.c_str(), L[i].chain.c_str(), L[i].rnum, L[i].axyz[0], L[i].axyz[1], L[i].axyz[2]);
    }
    fclose(pdb_file);
}

/**
 * point_transform:
 * @p: a Point.
 * @m: the Matrix representing the transformation to 
 * @origin: orgin of the reference frame
 * apply to the coordinates of @p.
 *
 * Transform the coordinates of @p according to @m. (p[] becomes m[][].p[]).
 */


void point_transform (double* p, T43Matrix m)
{
    double x, y, z;
    x = m[0][0]*p[0] + m[0][1]*p[1] + m[0][2]*p[2] + m[3][0];
    y = m[1][0]*p[0] + m[1][1]*p[1] + m[1][2]*p[2] + m[3][1];
    z = m[2][0]*p[0] + m[2][1]*p[1] + m[2][2]*p[2] + m[3][2];
    p[0] = x; p[1] = y; p[2] = z;
    assert(!isnan(x));
    assert(!isnan(y));
    assert(!isnan(z));
}

// tranform coordinates for the ligand
void transform(T43Matrix R, vector<atom>& ain, vector<atom>& atrans)
{
    for(size_t i=0;i<ain.size();i++)
    {
        atom t = ain[i];
        point_transform(t.axyz, R);
        atrans.push_back(t);
    }
}


void rotate_coord(double oldX, double oldY, double oldZ,
                 double& newX, double& newY, double& newZ,
                 double psi, double theta, double phi)
{
    double r11, r21, r31, r12, r22, r32, r13, r23, r33;

    r11 = cos(psi)*cos(phi)-cos(theta)*sin(phi)*sin(psi);
    r12 = cos(psi)*sin(phi) + cos(theta)*cos(phi)*sin(psi);
    r13 = sin(psi)*sin(theta);
    r21 = -sin(psi)*cos(phi) - cos(theta)*sin(phi)*cos(psi);
    r22 = -sin(psi)*sin(phi) + cos(theta)*cos(phi)*cos(psi);
    r23 = cos(psi)*sin(theta);
    r31 = sin(theta)*sin(phi);
    r32 = -sin(theta)*cos(phi);
    r33 = cos(theta);

    newX = r11 * oldX + r12 * oldY + r13 * oldZ;
    newY = r21 * oldX + r22 * oldY + r23 * oldZ;
    newZ = r31 * oldX + r32 * oldY + r33 * oldZ;
}

void apply_rotation(vector<atom>& P, double* rot)
{
    double tx1, ty1, tz1, oldx, oldy, oldz;
    double rand1 = rot[0], rand2 = rot[1], rand3 = rot[2];
    vector<atom> NB;
    for(size_t i=0;i<P.size();i++)
    {
        atom RA = P[i];
        oldx = RA.axyz[0];
        oldy = RA.axyz[1];
        oldz = RA.axyz[2];

        rotate_coord(oldx, oldy, oldz, tx1, ty1, tz1, rand1, rand2, rand3);
        RA.axyz[0] = tx1;RA.axyz[1] = ty1;RA.axyz[2] = tz1;

        NB.push_back(RA);
    }
    P.clear();
    P = NB;
}

void translate_atoms(vector<atom>& C, double* cog)
{
    for(size_t i=0;i<C.size();i++)
    {
        for(int j=0;j<3;j++)
        {
             C[i].axyz[j] += cog[j];
        }
    }
}

int main(int argc, char** argv)
{
    if(argv[1] == NULL || argv[2] == NULL || argv[3] == NULL || argv[4] == NULL || argv[5] == NULL)
    {
        cerr << "Usage: PDBGEN receptor.pdb ligand.pdb prediction_file number_of_predictions_to_output outfilename" << endl;
        exit(EXIT_FAILURE);
    }

    string rec(argv[1]);    
    // read pdb coordinates and atom info for first protein
    vector<atom> R;
    read_protein(rec, R);

    // read pdb coordinates and atom info for second protein
    string lig(argv[2]);    
    vector<atom> L;
    read_protein(lig, L);

    string predfile(argv[3]);
    int num_pred = atoi(argv[4]);
    string fname = argv[5];

    vector<prediction> Z;
    double rotL[3], transL[3];
    read_predictions(predfile, Z, transL, rotL);
    
    translate_atoms(L, transL);
    apply_rotation(L, rotL);

    sort(Z.begin(), Z.end(), scoresort);

    if(num_pred > (int)Z.size())
    {
        num_pred = (int)Z.size();
    }

    for(int i=0;i<num_pred;i++)
    {
        T43Matrix M;
		int k = 0;
    	for(int l=0;l<3;l++)
    	{
            for(int j=0;j<3;j++)
            {
                M[l][j] = Z[i].T[k];
                k++;
            }
        }
    	M[3][0] = Z[i].T[9];
    	M[3][1] = Z[i].T[10];
    	M[3][2] = Z[i].T[11];
    	vector<atom> atrans;
	    transform(M, L, atrans);
        print_complex(fname, i+1, atrans);
    }
    
    exit(EXIT_SUCCESS);
}
